local Commands = require("utility/commands")

local colors = {
    white = {r = 1, g = 1, b = 1},
    black = {r = 0, g = 0, b = 0},
    darkgrey = {r = 0.25, g = 0.25, b = 0.25},
    grey = {r = 0.5, g = 0.5, b = 0.5},
    lightgrey = {r = 0.75, g = 0.75, b = 0.75},
    red = {r = 1, g = 0, b = 0},
    darkred = {r = 0.5, g = 0, b = 0},
    lightred = {r = 1, g = 0.5, b = 0.5},
    green = {r = 0, g = 1, b = 0},
    darkgreen = {r = 0, g = 0.5, b = 0},
    lightgreen = {r = 0.5, g = 1, b = 0.5},
    blue = {r = 0, g = 0, b = 1},
    darkblue = {r = 0, g = 0, b = 0.5},
    lightblue = {r = 0.5, g = 0.5, b = 1},
    orange = {r = 1, g = 0.55, b = 0.1},
    yellow = {r = 1, g = 1, b = 0},
    pink = {r = 1, g = 0, b = 1},
    purple = {r = 0.6, g = 0.1, b = 0.6},
    brown = {r = 0.6, g = 0.4, b = 0.1}
}

local function add_job(tick, job)
    if not global.watch_queue[tick] then
        global.watch_queue[tick] = {}
    end
    table.insert(global.watch_queue[tick], job)
end

local function InitState()
    global.watch_queue = global.watch_queue or {}
    global.tick_tock = global.tick_tock or "tick"
end

local function OnNthTick(e)
    InitState()

    local merge = {}
    local bomb = {}

    if global.watch_queue then
        -- Merge all bombs together, taking the max scheduled tick
        for schedule_tick, job in pairs(global.watch_queue) do
            for _, work in pairs(job) do
                if work and work.target then
                    if not merge[work.target] then
                        merge[work.target] = {delay = schedule_tick - e.tick, payload = work.payload}
                    else
                        merge[work.target] = {
                            delay = merge[work.target].delay + (schedule_tick - e.tick),
                            payload = merge[work.target].payload + work.payload
                        }
                    end
                    log("merge target (" .. work.target .. ") delay (" .. merge[work.target].delay .. ") payload (" .. merge[work.target].payload .. ")")
                end
            end
        end

        -- Put the merged results back into the queue
        global.watch_queue = {}
        for target, work in pairs(merge) do
            local p = game.get_player(target)
            local delay = e.tick + work.delay
            local msg = global.tick_tock -- .. " " .. math.floor(work.delay / 60)

            if p and p.character and (e.tick % 60) == 0 then
                if (e.tick % (60 * settings.global["jdgb_display_time"].value)) == 0 then
                    msg = global.tick_tock .. " " .. math.floor(work.delay / 60)
                end
                p.surface.create_entity {name = "flying-text", position = p.position, text = msg, color = colors.white}
            end

            if p and p.ticks_to_respawn then
                delay = delay + 20
            end

            add_job(delay, {target = target, payload = work.payload})
            log("reschedule (" .. target .. ") delay (" .. delay .. ") payload (" .. work.payload .. ")")

            if global.tick_tock == "tick" then
                global.tick_tock = "tock"
            else
                global.tick_tock = "tick"
            end
        end

        -- Now process the queue
        for schedule_tick, job in pairs(global.watch_queue) do
            for _, work in pairs(job) do
                if schedule_tick <= e.tick then
                    if not bomb[work.target] then
                        bomb[work.target] = 0
                    end
                    bomb[work.target] = bomb[work.target] + work.payload
                    global.watch_queue[schedule_tick] = nil
                    log("bomb (" .. work.target .. ") payload (" .. work.payload .. ")")
                end
            end
        end
    end

    for k, v in pairs(bomb) do
        local p = game.get_player(k)
        if not p then
            log("player (" .. k .. ") not present on server .. reschedule")
            add_job(e.tick + (60 * 60), {target = k, payload = v})
        elseif not p.character then -- if bomb trigged when character is dead reschedule
            log("player (" .. k .. ") does not have a character .. reschedule")
            local delay = 60 * 10
            add_job(e.tick + delay, {target = p.name, payload = v})
        else
            -- Snoso logic
            local a, b, c = 0.002381, 2.811, 4.781
            local new_value = a * (v) ^ b + c
            new_value = math.floor(new_value)
            -- Snoso logic

            -- Some logic to keep the number between 1 and 1000
            if new_value > settings.global["jdgb_max_boom"].value then
                new_value = settings.global["jdgb_max_boom"].value
            end
            if new_value < 5 then
                new_value = 1
            end

            log("old_value (" .. v .. ") new_value (" .. new_value .. ")")
            p.surface.create_entity {name = "JGB-" .. new_value, position = p.position, target = p.character, speed = 1, max_range = 2}
        end
    end
end

local function ScheduleNuke(e)
    local args = Commands.GetArgumentsFromCommand(e.parameter)
    if #args ~= 2 then
        return
    end

    local target = args[1]
    local size = tonumber(args[2])
    if size < 1 then
        return
    end

    local delay = 60 * size / 2
    delay = delay + e.tick

    add_job(delay, {target = target, payload = size})
    log("" .. e.tick .. " Nuke scheduled for tick (" .. delay .. ") payload (" .. size .. ")")
end

local function ScheduleNukeAll(e)
    local args = Commands.GetArgumentsFromCommand(e.parameter)
    if #args ~= 1 then
        return
    end

    for _, player in pairs(game.connected_players) do
        e.parameter = "" .. player.name .. " " .. args[1]
        ScheduleNuke(e)
    end
end

Commands.Register("jd_goes_boom", {"api-description.jd_goes_boom"}, ScheduleNuke, true)
Commands.Register("all_go_boom", {"api-description.jd_goes_boom"}, ScheduleNukeAll, true)

script.on_nth_tick(20, OnNthTick)
script.on_init(InitState)
script.on_configuration_changed(InitState)
